import argparse
import os
from typing import Dict, Union

from jina import __stop_msg__
from jina.peapods.runtimes import BaseRuntime
from jina.peapods.peas import BasePea

__all__ = ['LocalRuntime']


class LocalRuntime(BaseRuntime):

    """LocalRuntime is a process or thread providing the support to run different :class:`BasePea` locally.

        Inside the run method, the :class:`BasePea` is context managed to guarantee a robust closing of the Pea context
    """
    def __init__(self,
                 args: Union['argparse.Namespace', Dict],
                 pea_cls: 'BasePea' = BasePea):
        super().__init__(args)
        self._envs = {'JINA_POD_NAME': self.name,
                      'JINA_LOG_ID': self.args.log_id}

        if 'env' in self.args and self.args.env:
            self._envs.update(self.args.env)
        self.pea = pea_cls(self.args, ctrl_addr=self.ctrl_addr, ctrl_with_ipc=self.ctrl_with_ipc)

    def set_environment_vars(self):
        """Set environment variable to this pea

        .. note::
            Please note that env variables are process-specific. Subprocess inherits envs from
            the main process. But Subprocess's envs do NOT affect the main process. It does NOT
            mess up user local system envs.

        .. warning::
            If you are using ``thread`` as backend, envs setting will likely be overidden by others
        """
        if self._envs:
            if self.args.runtime == 'thread':
                self.logger.warning('environment variables should not be set when runtime="thread". '
                                    f'ignoring all environment variables: {self._envs}')
            else:
                for k, v in self._envs.items():
                    os.environ[k] = v

    def unset_environment_vars(self):
        if self._envs and self.args.runtime != 'thread':
            for k in self._envs.keys():
                os.unsetenv(k)

    @property
    def is_idle(self) -> bool:
        return self.pea.is_idle

    def run(self):
        """Start the request loop of this Runtime. It will start a BasePea as a context manager and call its
        main run entrypoint """
        try:
            self.set_environment_vars()
            with self.pea as pea:
                pea.run(self.is_ready_event)
        except Exception as ex:
            self.logger.info(f'runtime run caught {repr(ex)}')
        finally:
            # if an exception occurs this unsets ready and shutting down
            self.unset_environment_vars()
            self.unset_ready()
            self.logger.success(__stop_msg__)
            self.set_shutdown()



import argparse
from typing import Union, Dict

from jina.peapods.peas import PeaRoleType
from jina.peapods.zmq import Zmqlet, send_ctrl_message
from jina.peapods.runtimes import BaseRuntime

__all__ = ['BaseRemoteRuntime']


class BaseRemoteRuntime(BaseRuntime):
    """A RemoteRuntime that will spawn a remote `BasePea` or `BasePod`.

    Inside the run method, a remote container is started and its lifetime monitored and managed.

    Classes inheriting from it must inherit :meth:`_monitor_remote`

    """

    def __init__(self, args: Union['argparse.Namespace', Dict], kind: str):
        super().__init__(args)
        self.kind = kind
        self.all_ctrl_addr = []
        if isinstance(self.args, Dict):
            first_pea_args = self.args['peas'][0]
            self.ctrl_timeout = first_pea_args.timeout_ctrl
            self.daemon = first_pea_args.daemon
            if first_pea_args.name:
                self.name = first_pea_args.name
            if first_pea_args.role == PeaRoleType.PARALLEL:
                self.name = f'{self.name}-{first_pea_args.pea_id}'
            for args in self.args['peas']:
                ctrl_addr, _ = Zmqlet.get_ctrl_address(args.host, args.port_ctrl, args.ctrl_with_ipc)
                self.all_ctrl_addr.append(ctrl_addr)
        elif isinstance(self.args, argparse.Namespace):
            self.daemon = self.args.daemon
            self.all_ctrl_addr.append(self.ctrl_addr)

    @property
    def is_idle(self) -> bool:
        raise NotImplementedError

    def _monitor_remote(self):
        raise NotImplementedError

    def run(self):
        """Start the remote monitoring loop. Will spawn a BasePea in a remote machine.
         It will communicate with the remote to see when it is ready to receive messages from the rest
         of the flow and stream the logs from the remote BasePea"""
        try:
            self._monitor_remote()
        except KeyboardInterrupt:
            self.logger.info('Loop interrupted by user')
        except SystemError as ex:
            self.logger.error(f'SystemError interrupted pea loop {repr(ex)}')
        except Exception as ex:
            self.logger.critical(f'unknown exception: {repr(ex)}', exc_info=True)
        finally:
            self.logger.info(f'Ended monitoring remote')

    def send_terminate_signal(self) -> None:
        """Gracefully close this pea and release all resources """
        if self.is_ready_event.is_set() and self.all_ctrl_addr:
            for ctrl_addr in self.all_ctrl_addr:
                send_ctrl_message(address=ctrl_addr, cmd='TERMINATE',
                                  timeout=self.ctrl_timeout)

    def close(self) -> None:
        """Close this `Runtime` by sending a `terminate signal` to the managed `BasePea`.
         Instead of waiting for BasePea to be closed, force shutdown to be set so that monitoring
         ends.

        :note: In this case, there is no risk at killing this `Process` without waiting for `BasePea` to properly
        close since the `LocalRuntime` in the remote machine will ensure that the `BasePea` is properly closed.
         """
        # Needed to override from Runtime because this forces shutdown event to end monitoring remote
        self.send_terminate_signal()
        self.unset_ready()
        self.is_shutdown.set()
        self.logger.close()
        if not self.daemon:
            self.join()



import argparse
import multiprocessing
import threading
from multiprocessing.synchronize import Event
from typing import Dict, Union

from jina.peapods.zmq import send_ctrl_message, Zmqlet
from jina.enums import PeaRoleType
from jina.excepts import PeaFailToStart

from jina.helper import typename
from jina.logging import JinaLogger

__all__ = ['RuntimeMeta', 'BaseRuntime']


def _get_event(obj: 'BaseRuntime') -> Event:
    if isinstance(obj, threading.Thread):
        return threading.Event()
    elif isinstance(obj, multiprocessing.Process):
        return multiprocessing.Event()
    else:
        raise NotImplementedError


def _make_or_event(obj: 'BaseRuntime', *events) -> Event:
    or_event = _get_event(obj)

    def or_set(self):
        self._set()
        self.changed()

    def or_clear(self):
        self._clear()
        self.changed()

    def orify(e, changed_callback):
        e._set = e.set
        e._clear = e.clear
        e.changed = changed_callback
        e.set = lambda: or_set(e)
        e.clear = lambda: or_clear(e)

    def changed():
        bools = [e.is_set() for e in events]
        if any(bools):
            or_event.set()
        else:
            or_event.clear()

    for e in events:
        orify(e, changed)
    changed()
    return or_event


class RuntimeMeta(type):
    """Meta class of :class:`BaseRuntime` to enable switching between ``thread`` and ``process`` backend. """
    _dct = {}

    def __new__(cls, name, bases, dct):
        _cls = super().__new__(cls, name, bases, dct)
        RuntimeMeta._dct.update({name: {'cls': cls,
                                        'name': name,
                                        'bases': bases,
                                        'dct': dct}})
        return _cls

    def __call__(cls, *args, **kwargs) -> 'RuntimeMeta':
        # switch to the new backend
        _cls = {
            'thread': threading.Thread,
            'process': multiprocessing.Process,
        }.get(getattr(args[0], 'runtime', 'thread'))

        # rebuild the class according to mro
        for c in cls.mro()[-2::-1]:
            arg_cls = RuntimeMeta._dct[c.__name__]['cls']
            arg_name = RuntimeMeta._dct[c.__name__]['name']
            arg_dct = RuntimeMeta._dct[c.__name__]['dct']
            _cls = super().__new__(arg_cls, arg_name, (_cls,), arg_dct)

        return type.__call__(_cls, *args, **kwargs)


class BaseRuntime(metaclass=RuntimeMeta):

    """BaseRuntime is a process or thread providing the support to run different :class:`BasePea` in different environments.
    It manages the lifetime of these `BasePea` objects living in `Local`, `Remote`, or `Container` environment.

    Inherited classes must define their own `run` method that is the one that will be run in a separate process or thread than the main process
    """

    def __init__(self, args: Union['argparse.Namespace', Dict]):
        super().__init__()
        self.args = args
        self.name = self.__class__.__name__  #: this is the process name

        self.is_ready_event = _get_event(self)
        self.is_shutdown = _get_event(self)
        self.ready_or_shutdown = _make_or_event(self, self.is_ready_event, self.is_shutdown)
        self.is_shutdown.clear()

        if 'daemon' in args:
            self.daemon = args.daemon
        if 'name' in self.args and self.args.name:
            self.name = f'runtime-{self.args.name}'
        if 'role' in self.args and self.args.role == PeaRoleType.PARALLEL:
            self.name = f'runtime-{self.args.name}-{self.args.pea_id}'
        if 'role' in self.args and self.args.role == PeaRoleType.HEAD:
            self.name = f'runtime-{self.args.name}-head'
        if 'role' in self.args and self.args.role == PeaRoleType.TAIL:
            self.name = f'runtime-{self.args.name}-tail'
        if 'host' in self.args and 'port_ctrl' in self.args and 'ctrl_with_ipc' in self.args:
            self.ctrl_addr, self.ctrl_with_ipc = Zmqlet.get_ctrl_address(self.args.host, self.args.port_ctrl,
                                                                         self.args.ctrl_with_ipc)

        if 'log_id' in self.args and 'log_config' in self.args:
            self.logger = JinaLogger(self.name,
                                     log_id=self.args.log_id,
                                     log_config=self.args.log_config)
        else:
            self.logger = JinaLogger(self.name)

    def run(self):
        raise NotImplementedError

    def start(self):
        super().start()
        if isinstance(self.args, dict):
            _timeout = getattr(self.args['peas'][0], 'timeout_ready', -1)
        else:
            _timeout = getattr(self.args, 'timeout_ready', -1)

        if _timeout <= 0:
            _timeout = None
        else:
            _timeout /= 1e3

        if self.ready_or_shutdown.wait(_timeout):
            if self.is_shutdown.is_set():
                # return too early and the shutdown is set, means something fails!!
                self.logger.critical(f'fails to start {typename(self)} with name {self.name}, '	
                                     f'this often means the executor used in the pod is not valid')
                raise PeaFailToStart
            else:
                self.logger.info(f'ready to listen')
            return self
        else:
            raise TimeoutError(
                f'{typename(self)} with name {self.name} can not be initialized after {_timeout * 1e3}ms')

    def set_ready(self):
        """Set the `is_ready_event` to indicate that the `BasePea` managed by the Runtime is ready to start
         receiving messages"""
        self.is_ready_event.set()

    def unset_ready(self):
        """Clear the `is_ready_event` to indicate that the `BasePea` managed by the Runtime is not anymore ready to start
         receiving messages"""
        self.is_ready_event.clear()

    def set_shutdown(self):
        """Set the `is_shutdown` event to indicate that the `BasePea` managed by the Runtime is closed and the parallel process
        can be shutdown"""
        self.is_shutdown.set()

    @property
    def status(self):
        """Send the control signal ``STATUS`` to the manages `BasePea` and return the status """
        return send_ctrl_message(self.ctrl_addr, 'STATUS', timeout=self.args.timeout_ctrl)

    @property
    def is_ready(self) -> bool:
        status = self.status
        return status and status.is_ready

    @property
    def is_idle(self) -> bool:
        raise NotImplementedError

    def send_terminate_signal(self):
        """Send a terminate signal to the `BasePea` supported by this `Runtime` """
        return send_ctrl_message(self.ctrl_addr, 'TERMINATE', timeout=self.args.timeout_ctrl)

    def close(self) -> None:
        """Close this `Runtime` by sending a `terminate signal` to the managed `BasePea`. Wait to
         be sure that the `BasePea` is properly closed to join the parallel process """
        self.send_terminate_signal()
        self.is_shutdown.wait()
        self.logger.close()
        if not self.daemon:
            self.join()

    def __enter__(self):
        return self.start()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional

from .. import __default_host__
from ..enums import RemoteAccessType
from ..helper import is_valid_local_config_source
from ..logging import default_logger
from .peas import BasePea

if False:
    import argparse


def Runtime(args: Optional['argparse.Namespace'] = None,
            allow_remote: bool = True,
            pea_cls: BasePea = BasePea, **kwargs):
    """Initialize a :class:`LocalRuntime`, :class:`ContainerRuntime` or :class:`RemoteRuntime`

    :param args: arguments from CLI
    :param allow_remote: allow start a :class:`RemoteRuntime`
    :param pea_cls: declares the type of `Pea` to be instantiated by `LocalRuntime`
    :param kwargs: all supported arguments from CLI

    """
    if args is None:
        from ..parser import set_pea_parser
        from ..helper import get_parsed_args
        _, args, _ = get_parsed_args(kwargs, set_pea_parser())
    if not allow_remote:
        # set the host back to local, as for the remote, it is running "locally"
        if args.host != __default_host__:
            args.host = __default_host__
            default_logger.warning(f'setting host to {__default_host__} as allow_remote set to False')

    if args.host != __default_host__:
        from .runtimes.remote.jinad import JinadRemoteRuntime
        return JinadRemoteRuntime(args, kind='pea')
    elif args.uses and not is_valid_local_config_source(args.uses):
        from .runtimes.container import ContainerRuntime
        return ContainerRuntime(args)
    else:
        from .runtimes.local import LocalRuntime
        return LocalRuntime(args, pea_cls=pea_cls)


def Pod(args: Optional['argparse.Namespace'] = None,
        allow_remote: bool = True, **kwargs):
    """Initialize a :class:`BasePod`, :class:`JinadRemoteRuntime`, :class:`MutablePod` or :class:`SSHRuntime`

    :param args: arguments from CLI
    :param allow_remote: allow start a :class:`RemoteSSHPod`
    :param kwargs: all supported arguments from CLI
    """
    if args is None:
        from ..parser import set_pod_parser
        from ..helper import get_parsed_args
        _, args, _ = get_parsed_args(kwargs, set_pod_parser())
    if isinstance(args, dict):
        hosts = set()
        for k in args.values():
            if k:
                if not isinstance(k, list):
                    k = [k]
                for kk in k:
                    if not allow_remote and kk.host != __default_host__:
                        kk.host = __default_host__
                        default_logger.warning(f'host is reset to {__default_host__} as allow_remote=False')
                    hosts.add(kk.host)

        if len(hosts) == 1:
            if __default_host__ in hosts:
                from .pods.mutable import MutablePod
                return MutablePod(args)
            else:
                from .runtimes.remote.jinad import JinadRemoteRuntime
                return JinadRemoteRuntime(args, kind='pod')

    if not allow_remote and args.host != __default_host__:
        args.host = __default_host__
        default_logger.warning(f'host is reset to {__default_host__} as allow_remote=False')

    if args.host != __default_host__:
        if args.remote_access == RemoteAccessType.JINAD:
            from .runtimes.remote.jinad import JinadRemoteRuntime
            return JinadRemoteRuntime(args, kind='pod')
        elif args.remote_access == RemoteAccessType.SSH:
            from .runtimes.remote.ssh import SSHRuntime
            return SSHRuntime(args, kind='pod')
        else:
            raise ValueError(f'{args.remote_access} is not supported')
    else:
        from .pods import BasePod
        return BasePod(args)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Sequence, Any

import numpy as np

from ..executors.encoders import BaseImageEncoder
from ..executors.evaluators.rank import BaseRankingEvaluator


class MyEncoder(BaseImageEncoder):
    """Simple Encoder used in :command:`jina hello-world`,
        it transforms the original 784-dim vector into a 64-dim vector using
        a random orthogonal matrix, which is stored and shared in index and query time"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        np.random.seed(1337)
        # generate a random orthogonal matrix
        H = np.random.rand(784, 64)
        u, s, vh = np.linalg.svd(H, full_matrices=False)
        self.oth_mat = u @ vh
        self.touch()

    def encode(self, data: 'np.ndarray', *args, **kwargs):
        # reduce dimension to 50 by random orthogonal projection
        return (data.reshape([-1, 784]) / 255) @ self.oth_mat


class MyEvaluator(BaseRankingEvaluator):
    """MyEvaluator differs from :class:`PrecisionEvaluator` because it evaluates how many of the returned matches
    are of the same label as the true nearest neighbor
    """

    @property
    def metric(self):
        return f'LabelPrecision@{self.eval_at}'

    def evaluate(self, actual: Sequence[Any], desired: Sequence[Any], *args, **kwargs) -> float:
        desired_label = desired[0]

        ret = 0.0
        for match_label in actual[:self.eval_at]:
            if match_label == desired_label:
                ret += 1.0

        divisor = min(self.eval_at, len(actual))
        return ret / divisor if divisor != 0.0 else 0.0



from argparse import Namespace
from typing import Union, Dict, Optional

from jina.peapods.runtimes.remote.jinad.api import PeaJinadAPI
from jina.peapods.runtimes.remote import BaseRemoteRuntime
from jina.helper import cached_property, namespace_to_dict, colored, typename
from .api import get_jinad_api


class JinadRemoteRuntime(BaseRemoteRuntime):
    """A JinadRemoteRuntime that will spawn a remote `BasePea` or `BasePod` via REST communication with a jinad instance
    """
    def __init__(self, args: Union['Namespace', Dict], kind: str):
        super().__init__(args, kind=kind)
        if isinstance(self.args, Namespace):
            self.ctrl_timeout = self.args.timeout_ctrl
        if isinstance(self.args, Dict):
            api_args = self.args['peas'][0]
        else:
            api_args = self.args

        self.api = get_jinad_api(kind=self.kind, host=api_args.host, port=api_args.port_expose, logger=self.logger)

    @cached_property
    def remote_id(self) -> Optional[str]:
        return self.spawn_remote()

    def spawn_remote(self, **kwargs) -> Optional[str]:
        if self.api.is_alive:
            args = namespace_to_dict(self.args)
            if self.api.upload(args, **kwargs):
                return self.api.create(args, **kwargs)

    def _monitor_remote(self):
        if self.remote_id:
            self.logger.success(f'created remote {self.api.kind} with id {colored(self.remote_id, "cyan")}')
            self.set_ready()
            self.api.log(self.remote_id, self.is_shutdown)
        else:
            self.logger.error(f'fail to create {typename(self)} remotely')



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import argparse
import asyncio
import os
import time
from pathlib import Path
from typing import Dict, Union

from jina.peapods.zmq import Zmqlet
from jina import __ready_msg__, __stop_msg__
from jina.helper import is_valid_local_config_source, kwargs2list, get_non_defaults_args
from jina.logging import JinaLogger
from jina.peapods.runtimes import BaseRuntime

__all__ = ['ContainerRuntime']


class ContainerRuntime(BaseRuntime):
    """A ContainerRuntime that will spawn a dockerized `BasePea`. It requires a docker-corresponding valid ``args.uses``

    Inside the run method, a docker container is started in the same host where this instance lives, and its logs and lifetime
    is controlled.

    """

    def __init__(self, args: Union['argparse.Namespace', Dict]):
        super().__init__(args)
        import docker
        # recompute the control_addr, do not assign client, since this would be an expensive object to
        # copy in the new process generated
        client = docker.from_env()

        from sys import platform
        # Related to potential docker-in-docker communication. If `ContainerPea` lives already inside a container.
        # it will need to communicate using the `bridge` network.
        if platform in ('linux', 'linux2'):
            try:
                bridge_network = client.networks.get('bridge')
                if bridge_network:
                    self.ctrl_addr, self.ctrl_with_ipc = Zmqlet.get_ctrl_address(
                        bridge_network.attrs['IPAM']['Config'][0]['Gateway'],
                        self.args.port_ctrl,
                        self.args.ctrl_with_ipc)
            except Exception as exc:
                self.logger.warning(f'Unable to set control address from "bridge" network: {repr(exc)}'
                                    f' Control address set to {self.ctrl_addr}')
            finally:
                pass
        client.close()

    @property
    def is_idle(self) -> bool:
        raise NotImplementedError

    def _docker_run(self):
        # important to notice, that client is not assigned as instance member to avoid potential
        # heavy copy into new process memory space
        import docker
        client = docker.from_env()

        # the image arg should be ignored otherwise it keeps using ContainerPea in the container
        # basically all args in BasePea-docker arg group should be ignored.
        # this prevent setting containerPea twice
        from jina.parser import set_pea_parser
        non_defaults = get_non_defaults_args(self.args, set_pea_parser(),
                                             taboo={'uses', 'entrypoint', 'volumes', 'pull_latest'})

        if self.args.pull_latest:
            self.logger.warning(f'pulling {self.args.uses}, this could take a while. if you encounter '
                                f'timeout error due to pulling takes to long, then please set '
                                f'"timeout-ready" to a larger value.')
            client.images.pull(self.args.uses)

        _volumes = {}
        if self.args.uses_internal:
            if os.path.exists(self.args.uses_internal):
                # external YAML config, need to be volumed into the container
                # uses takes value from uses_internal
                non_defaults['uses'] = '/' + os.path.basename(self.args.uses_internal)
                _volumes[os.path.abspath(self.args.uses_internal)] = {'bind': non_defaults['uses'], 'mode': 'ro'}
            elif not is_valid_local_config_source(self.args.uses_internal):
                raise FileNotFoundError(
                    f'"uses_internal" {self.args.uses_internal} is not like a path, please check it')
        if self.args.volumes:
            for p in self.args.volumes:
                Path(os.path.abspath(p)).mkdir(parents=True, exist_ok=True)
                _p = '/' + os.path.basename(p)
                _volumes[os.path.abspath(p)] = {'bind': _p, 'mode': 'rw'}

        _expose_port = [self.args.port_ctrl]
        if self.args.socket_in.is_bind:
            _expose_port.append(self.args.port_in)
        if self.args.socket_out.is_bind:
            _expose_port.append(self.args.port_out)

        from sys import platform
        if platform in ('linux', 'linux2'):
            net_mode = 'host'
        else:
            net_mode = None

        _args = kwargs2list(non_defaults)
        ports = {f'{v}/tcp': v for v in _expose_port} if not net_mode else None
        self._container = client.containers.run(self.args.uses, _args,
                                                detach=True,
                                                auto_remove=True,
                                                ports=ports,
                                                name=self.name,
                                                volumes=_volumes,
                                                network_mode=net_mode,
                                                entrypoint=self.args.entrypoint)

        # wait until the container is ready
        self.logger.info('waiting ready signal from the container')
        client.close()

    def _monitor_pea_in_container(self):
        """Direct the log from the container to local console """

        def check_ready():
            while not self.is_ready:
                time.sleep(0.1)
            self.is_ready_event.set()
            self.logger.success(__ready_msg__)
            return True

        async def _loop_body():
            import docker
            loop = asyncio.get_event_loop()
            loop.run_in_executor(None, check_ready)

            logger = JinaLogger('🐳', **vars(self.args))

            with logger:
                try:
                    for line in self._container.logs(stream=True):
                        logger.info(line.strip().decode())
                except docker.errors.NotFound:
                    self.logger.error('the container can not be started, check your arguments, entrypoint')

        asyncio.run(_loop_body())

    def run(self):
        """Start the container loop. Will spawn a docker container with a BasePea running inside.
         It will communicate with the container to see when it is ready to receive messages from the rest
         of the flow and stream the logs from the pea in the container"""
        try:
            self._docker_run()
            self._monitor_pea_in_container()
        except KeyboardInterrupt:
            self.logger.info('Loop interrupted by user')
        except SystemError as ex:
            self.logger.error(f'SystemError interrupted pea loop {repr(ex)}')
        except Exception as ex:
            self.logger.critical(f'unknown exception: {repr(ex)}', exc_info=True)
        finally:
            from docker.errors import NotFound
            try:
                if getattr(self, '_container'):
                    self._container.stop()
            except NotFound:
                self.logger.warning(
                    'the container is already shutdown (mostly because of some error inside the container)')
            self.unset_ready()
            self.logger.success(__stop_msg__)
            self.set_shutdown()



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import gzip
import os
import urllib.request
import webbrowser
import random

import numpy as np
from pkg_resources import resource_filename

from ..helper import colored
from ..logging import default_logger
from ..logging.profile import ProgressBar
from .. import Document

result_html = []
num_docs_evaluated = 0
evaluation_value = 0.0


def compute_mean_evaluation(resp):
    global num_docs_evaluated
    global evaluation_value
    for d in resp.search.docs:
        num_docs_evaluated += 1
        evaluation_value += d.evaluations[0].value


def evaluate_generator(num_docs: int, target: dict):
    for j in range(num_docs):
        num_data = len(target['query-labels']['data'])
        n = random.randint(0, num_data)
        label_int = target['query-labels']['data'][n][0]
        document = Document(content=(target['query']['data'][n]))
        document.tags['label_id'] = str(label_int)
        ground_truth = Document()
        match = Document()
        match.tags['label_id'] = str(label_int)
        ground_truth.matches.append(match)
        yield document, ground_truth


def index_generator(num_docs: int, target: dict):
    for j in range(num_docs):
        d = Document(content=target['index']['data'][j])
        d.update_id()
        label_int = target['index-labels']['data'][j][0]
        d.tags['label_id'] = str(label_int)
        yield d


def query_generator(num_docs: int, target: dict):
    for n in range(num_docs):
        num_data = len(target['query-labels']['data'])
        n = random.randint(0, num_data)
        d = Document(content=(target['query']['data'][n]))
        d.update_id()
        label_int = target['query-labels']['data'][n][0]
        d.tags['label_id'] = str(label_int)
        yield d


def print_result(resp):
    for d in resp.search.docs:
        vi = d.uri
        result_html.append(f'<tr><td><img src="{vi}"/></td><td>')
        for kk in d.matches:
            kmi = kk.uri
            result_html.append(f'<img src="{kmi}" style="opacity:{kk.score.value}"/>')
        result_html.append('</td></tr>\n')


def write_html(html_path):
    with open(resource_filename('jina', '/'.join(('resources', 'helloworld.html'))), 'r') as fp, \
            open(html_path, 'w') as fw:
        t = fp.read()
        t = t.replace('{% RESULT %}', '\n'.join(result_html))
        evaluation_percentage = evaluation_value/num_docs_evaluated * 100.0
        t = t.replace('{% EVALUATION %}', '{:.2f}%'.format(evaluation_percentage))
        fw.write(t)

    url_html_path = 'file://' + os.path.abspath(html_path)

    try:
        webbrowser.open(url_html_path, new=2)
    except:
        pass  # intentional pass, browser support isn't cross-platform
    finally:
        default_logger.success(f'You should see a "hello-world.html" opened in your browser, '
                               f'if not you may open {url_html_path} manually')

    colored_url = colored('https://opensource.jina.ai', color='cyan', attrs='underline')
    default_logger.success(
        f'🤩 Intrigued? Play with "jina hello-world --help" and learn more about Jina at {colored_url}')


def download_data(targets, download_proxy=None):
    opener = urllib.request.build_opener()
    if download_proxy:
        proxy = urllib.request.ProxyHandler({'http': download_proxy, 'https': download_proxy})
        opener.add_handler(proxy)
    urllib.request.install_opener(opener)
    with ProgressBar(task_name='download fashion-mnist', batch_unit='') as t:
        for k, v in targets.items():
            if not os.path.exists(v['filename']):
                urllib.request.urlretrieve(v['url'], v['filename'], reporthook=lambda *x: t.update(1))
            if k == 'index-labels' or k == 'query-labels':
                v['data'] = load_labels(v['filename'])
            if k == 'index' or k == 'query':
                v['data'] = load_mnist(v['filename'])


def load_mnist(path):
    with gzip.open(path, 'rb') as fp:
        return np.frombuffer(fp.read(), dtype=np.uint8, offset=16).reshape([-1, 784])


def load_labels(path: str):
    with gzip.open(path, 'rb') as fp:
        return np.frombuffer(fp.read(), dtype=np.uint8, offset=8).reshape([-1, 1])



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import argparse
import os
import time
from collections import defaultdict
from contextlib import ExitStack
from typing import Dict, Optional, Union, List
from multiprocessing.synchronize import Event

import zmq

from ..zmq import ZmqStreamlet
from ... import Message
from ... import Request
from ...enums import PeaRoleType, SkipOnErrorType
from ...excepts import RequestLoopEnd, NoExplicitMessage, ExecutorFailToLoad, MemoryOverHighWatermark, DriverError, \
    ChainedPodException
from ...executors import BaseExecutor
from ...helper import is_valid_local_config_source
from ...logging import JinaLogger
from ...logging.profile import used_memory, TimeDict
from ...proto import jina_pb2

__all__ = ['BasePea']


class BasePea(ExitStack):
    """BasePea is an unary service unit which provides network interface and
    communicates with others via protobuf and ZeroMQ. It also is a context manager of an Executor .
    """

    def __init__(self, args: Union['argparse.Namespace', Dict], **kwargs):
        """ Create a new :class:`BasePea` object

        :param args: the arguments received from the CLI
        """
        super().__init__()
        self.args = args
        self.name = self.__class__.__name__  #: this is the process name

        self.last_active_time = time.perf_counter()
        self.last_dump_time = time.perf_counter()

        self._timer = TimeDict()

        self._request = None
        self._message = None

        # all pending messages collected so far, key is the request id
        self._pending_msgs = defaultdict(list)  # type: Dict[str, List['Message']]
        self._partial_requests = None
        self._partial_messages = None

        # TODO: BAD pattern here!
        #   It is unnecessary to do "`a` in b and b.a" for `Namespace`, the only reason
        #   for this check is because Client is using (abusing?) `BasePea` but PyCilent's args is not
        #   generated by set_pea/pod_parser(). Therefore many information are missing and make this check
        #   necessary.
        #   (Han 2020.12.12)
        if 'daemon' in args:
            self.daemon = args.daemon
        if 'name' in self.args and self.args.name:
            self.name = self.args.name
        if 'role' in self.args and self.args.role == PeaRoleType.PARALLEL:
            self.name = f'{self.name}-{self.args.pea_id}'
        if 'log_id' in self.args and 'log_config' in self.args:
            self.logger = JinaLogger(self.name,
                                     log_id=self.args.log_id,
                                     log_config=self.args.log_config)
        else:
            self.logger = JinaLogger(self.name)

    def __str__(self):
        r = self.name
        if getattr(self, 'executor', None):
            r += f'({str(self.executor)})'
        return r

    def handle(self, msg: 'Message') -> 'BasePea':
        """Register the current message to this pea, so that all message-related properties are up-to-date, including
        :attr:`request`, :attr:`prev_requests`, :attr:`message`, :attr:`prev_messages`. And then call the executor to handle
        this message if its envelope's  status is not ERROR, else skip handling of message.

        :param msg: the message received
        """

        if self.expect_parts > 1 and self.expect_parts > len(self.partial_requests):
            # NOTE: reduce priority is higher than chain exception
            # otherwise a reducer will lose its function when eailier pods raise exception
            raise NoExplicitMessage

        if msg.envelope.status.code != jina_pb2.StatusProto.ERROR or self.args.skip_on_error < SkipOnErrorType.HANDLE:
            self.executor(self.request_type)
        else:
            raise ChainedPodException
        return self

    @property
    def is_idle(self) -> bool:
        """Return ``True`` when current time is ``max_idle_time`` seconds late than the last active time"""
        return (time.perf_counter() - self.last_active_time) > self.args.max_idle_time

    @property
    def request(self) -> 'Request':
        """Get the current request body inside the protobuf message"""
        return self._request

    @property
    def message(self) -> 'Message':
        """Get the current protobuf message to be processed"""
        return self._message

    @property
    def request_type(self) -> str:
        """Get the type of message being processed"""
        return self._message.envelope.request_type

    def load_executor(self):
        """Load the executor to this BasePea, specified by ``uses`` CLI argument.

        """
        try:
            self.executor = BaseExecutor.load_config(
                self.args.uses if is_valid_local_config_source(self.args.uses) else self.args.uses_internal,
                self.args.separated_workspace, self.args.pea_id, self.args.read_only)
            self.executor.attach(pea=self)
        except FileNotFoundError as ex:
            self.logger.error(f'fail to load file dependency: {repr(ex)}')
            raise ExecutorFailToLoad from ex
        except Exception as ex:
            raise ExecutorFailToLoad from ex

    def print_stats(self):
        self.logger.info(
            ' '.join(f'{k}: {v / self._timer.accum_time["loop"]:.2f}' for k, v in self._timer.accum_time.items()))

    def save_executor(self):
        """Save the contained executor according to the `dump_interval` parameter
        """
        if (time.perf_counter() - self.last_dump_time) > self.args.dump_interval > 0:
            self.executor.save()
            self.last_dump_time = time.perf_counter()
            if hasattr(self, 'zmqlet'):
                self.zmqlet.print_stats()

    @property
    def expect_parts(self) -> int:
        """The expected number of partial messages before trigger :meth:`handle` """
        return self.args.num_part if self.message.is_data_request else 1

    @property
    def partial_requests(self) -> List['Request']:
        """The collected partial requests under the current ``request_id`` """
        return self._partial_requests

    @property
    def partial_messages(self) -> List['Message']:
        """The collected partial messages under the current ``request_id`` """
        return self._partial_messages

    def pre_hook(self, msg: 'Message') -> 'BasePea':
        """Pre-hook function, what to do after first receiving the message """
        msg.add_route(self.name, self.args.identity)
        self._request = msg.request
        self._message = msg

        part_str = ' '
        if self.expect_parts > 1:
            req_id = msg.envelope.request_id
            self._pending_msgs[req_id].append(msg)
            self._partial_messages = self._pending_msgs[req_id]
            self._partial_requests = [v.request for v in self._partial_messages]
            part_str = f' ({len(self.partial_requests)}/{self.expect_parts} parts) '

        self.logger.info(f'recv {msg.envelope.request_type}{part_str}from {msg.colored_route}')
        return self

    def post_hook(self, msg: 'Message') -> 'BasePea':
        """Post-hook function, what to do before handing out the message """
        self.last_active_time = time.perf_counter()
        self.save_executor()
        self.check_memory_watermark()

        if self.expect_parts > 1:
            msgs = self._pending_msgs.pop(msg.envelope.request_id)
            msg.merge_envelope_from(msgs)

        msg.update_timestamp()
        return self

    def _callback(self, msg: 'Message'):
        self.is_post_hook_done = False  #: if the post_hook is called
        self.pre_hook(msg).handle(msg).post_hook(msg)
        self.is_post_hook_done = True
        return msg

    def _teardown(self):
        self.close_zmqlet()

    def msg_callback(self, msg: 'Message') -> None:
        """Callback function after receiving the message

        When nothing is returned then nothing is send out via :attr:`zmqlet.sock_out`.
        """
        try:
            # notice how executor related exceptions are handled here
            # generally unless executor throws an OSError, the exception are caught and solved inplace
            self.zmqlet.send_message(self._callback(msg))
        except RequestLoopEnd as ex:
            # this is the proper way to end when a terminate signal is sent
            self.logger.info(f'Terminating loop requested by terminate signal {repr(ex)}')
            self.zmqlet.send_message(msg)
            self._teardown()
        except (SystemError, zmq.error.ZMQError, KeyboardInterrupt) as ex:
            # save executor
            self.logger.info(f'{repr(ex)} causes the breaking from the event loop')
            self.zmqlet.send_message(msg)
            self._teardown()
        except MemoryOverHighWatermark:
            self.logger.critical(
                f'memory usage {used_memory()} GB is above the high-watermark: {self.args.memory_hwm} GB')
        except NoExplicitMessage:
            # silent and do not propagate message anymore
            # 1. wait partial message to be finished
            # 2. dealer send a control message and no need to go on
            pass
        except (RuntimeError, Exception, ChainedPodException) as ex:
            # general runtime error and nothing serious, we simply mark the message to error and pass on
            if not self.is_post_hook_done:
                self.post_hook(msg)
            if isinstance(ex, ChainedPodException):
                msg.add_exception()
                self.logger.warning(repr(ex))
            else:
                msg.add_exception(ex, executor=getattr(self, 'executor'))
                self.logger.error(repr(ex))
            if 'JINA_RAISE_ERROR_EARLY' in os.environ:
                raise
            self.zmqlet.send_message(msg)

    def request_loop(self, is_ready_event: 'Event'):
        """The body of the request loop
        """
        self.zmqlet = ZmqStreamlet(self.args, logger=self.logger)
        is_ready_event.set()
        self.zmqlet.start(self.msg_callback)

    def load_plugins(self):
        """Loads the plugins if needed necessary to load executors
        """
        if self.args.py_modules:
            from ...importer import PathImporter
            PathImporter.add_modules(*self.args.py_modules)

    def close_zmqlet(self):
        """Close the zmqlet if exists"""
        if hasattr(self, 'zmqlet'):
            self.zmqlet.close()

    def _initialize_executor(self):
        try:
            self.load_plugins()
            self.load_executor()
            return self.executor
        except Exception as ex:
            self.logger.critical(f'can not start a executor from {self.args.uses}', exc_info=True)
            raise ex

    def run(self, is_ready_event: 'Event'):
        """Start the request loop of this BasePea. It will listen to the network protobuf message via ZeroMQ. """
        try:
            # Every logger created in this process will be identified by the `Pod Id` and use the same name
            self.request_loop(is_ready_event)
        except KeyboardInterrupt:
            self.logger.info('Loop interrupted by user')
        except SystemError as ex:
            self.logger.error(f'SystemError interrupted pea loop {repr(ex)}')
        except DriverError as ex:
            self.logger.critical(f'driver error: {repr(ex)}', exc_info=True)
        except zmq.error.ZMQError:
            self.logger.critical('zmqlet can not be initiated')
        except Exception as ex:
            # this captures the general exception from the following places:
            # - self.zmqlet.recv_message
            # - self.zmqlet.send_message
            self.logger.critical(f'unknown exception: {repr(ex)}', exc_info=True)
        finally:
            self.logger.info(f'request loop ended, tearing down ...')
            self._teardown()

    def check_memory_watermark(self):
        """Check the memory watermark """
        if used_memory() > self.args.memory_hwm > 0:
            raise MemoryOverHighWatermark

    def __enter__(self) -> 'BasePea':
        executor = self._initialize_executor()

        if executor:
            self.enter_context(executor)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        super().__exit__(exc_type, exc_val, exc_tb)
        self._teardown()


def Pea(args: Optional['argparse.Namespace'] = None,
        gateway: bool = False,
        rest_api: bool = False,
        **kwargs):
    """Initialize a :class:`BasePea`, :class:`HeadPea`, :class:`TailPea`, or :class:`GatewayPea` or :class: `RESTGatewayPea`

    :param args: arguments from CLI
    :param gateway: true if gateway pea to be instantiated
    :param rest_api: true if gateway pea to be instantiated with REST (only considered if gateway is True)

    """
    if gateway:
        from .gateway import GatewayPea, RESTGatewayPea
        return RESTGatewayPea(args, **kwargs) if rest_api else GatewayPea(args, **kwargs)
    elif args.role == PeaRoleType.HEAD:
        from .headtail import HeadPea
        return HeadPea(args)
    elif args.role == PeaRoleType.TAIL:
        from .headtail import TailPea
        return TailPea(args)
    else:
        return BasePea(args)
